import argparse
from .decoder.FileDecoder import FileDecoder
import geojson
import json
import sys

def main():
    parser = argparse.ArgumentParser(description="Convert Vector Tile Protobuf files to GeoJSON, and saves it to a *.json file.")
    parser.add_argument("-i", "--input", required=True, dest='input_file', help="Input file")
    parser.add_argument("-o", "--output", required=True, dest='output_file',help="Output file")

    parser.add_argument("-x", dest = "tile_x", help="X coordinate of tile", required=False, type=int)
    parser.add_argument("-y", dest = "tile_y", help="Y coordinate of tile", required=False, type=int)
    parser.add_argument("-z", dest = "tile_z", help="Zoom level of tile", required=False, type=int)
    parser.add_argument("--json-indent", dest="json_indent", help="JSON file indentation. 0 or negative numbers generate dense JSON file.", default=0, type=int)
    parser.add_argument("--layer", dest = "layer", help="Only decode layer with given name. Outputs Pure GeoJSON.", required=False)
    parser.add_argument("--split-layers", dest = "split_layers", help="Split layers into separate GeoJSON files. Outputs Pure GeoJSON.", action="store_true", default=False)
    args = parser.parse_args()

    if(not (args.tile_x is not None and args.tile_y is not None and args.tile_z is not None)):
        print("Please provide tile coordinates.")
        exit(1)
    elif(args.output_file is None):
        print("No output file specified.")
    else:
        decoder = FileDecoder(args.tile_x, args.tile_y, args.tile_z, args.input_file)
        decoded = decoder.decode()
        if(args.layer is not None):
            if(args.layer in decoded):
                geojson.dump(decoded[args.layer], open(args.output_file, 'w'), indent=args.json_indent if args.json_indent > 0 else None)
                print("Writing layer {} to {}".format(args.layer, args.output_file))
            else:
                print("Layer %s not found in input file." % (args.layer))
        elif(args.split_layers):
            if(args.output_file.endswith(".json")):
                output_file_basename = args.output_file[:-5]
            else:
                output_file_basename = args.output_file
            
            for layer_name, layer_content in decoded.items():
                print("Writing layer {} to file {}.json".format(layer_name, output_file_basename + "-" + layer_name))
                json.dump(layer_content, open(output_file_basename + "-" + layer_name + ".json", 'w'), indent=args.json_indent if args.json_indent > 0 else None)
        else:
            json.dump(decoded, open(args.output_file, 'w'), indent=args.json_indent if args.json_indent > 0 else None)
            print("Wrote JSON to file {}".format(args.output_file))

if __name__ == '__main__':
    main()


from pyparsing import line
from . import vector_tile_pb2 as vt_proto
from geojson import Feature, Point, FeatureCollection, LineString, MultiLineString, MultiPolygon, Polygon, MultiPoint
from typing import Dict, Tuple, List
import math

def area_by_shoelace(points:  List[Tuple[float, float]]) -> float:
    "Assumes x,y points go around the polygon in one direction"
    x, y = zip(*points)
    return (sum(i * j for i, j in zip(x,             y[1:] + y[:1]))
           -sum(i * j for i, j in zip(x[1:] + x[:1], y            ))) / 2
               
def num2deg(xtile: int, ytile: int, zoom: int) -> Tuple[float, float]:
  n = 2.0 ** zoom
  lon_deg = xtile / n * 360.0 - 180.0
  lat_rad = math.atan(math.sinh(math.pi * (1 - 2 * ytile / n)))
  lat_deg = math.degrees(lat_rad)
  return (lat_deg, lon_deg)

def offset_to_latlon(xtile, ytile, zoom, xoffset, yoffset, extent):
    x_tiled_coord = xtile * extent + xoffset
    y_tiled_coord = ytile * extent + yoffset

    lon = x_tiled_coord * 360 / (extent * 2 ** zoom) - 180

    lat_noncoerce = 180 - y_tiled_coord * 360. / (extent * 2 ** zoom)
    lat = 360. / math.pi * math.atan(math.exp(lat_noncoerce * math.pi / 180)) - 90
    return (lon, lat)

def decode_file(filename: str, xtile, ytile, zoom) -> Dict[str, FeatureCollection]:
    tile = read_protobuf(filename)
    layers = tile.layers
    layers_dict = {}
    for layer in layers:
        layer_name, layer_content = extract_layer(layer, xtile, ytile, zoom)
        layers_dict[layer_name] = layer_content
    return layers_dict

def read_protobuf(filename: str) -> vt_proto.Tile:
    with open(filename, 'rb') as f:
        return vt_proto.Tile.FromString(f.read())

def unzigzag_coords(value: int) -> int:
    return (value >> 1) ^ (-(value & 1))

def expand_commands(cmds: List[int]) -> List[Tuple]:
    expanded_cmds = []
    cmds_pc = 0
    while(cmds_pc < len(cmds)):
        command_integer = cmds[cmds_pc]
        command_id = command_integer & 0x07
        command_count = command_integer >> 3
        if(command_id == 1 or command_id == 2):
            for i in range(command_count):
                cmds_pc += 1
                param1 = cmds[cmds_pc]
                cmds_pc += 1
                param2 = cmds[cmds_pc]
                expanded_cmds.append((command_id, param1, param2))
        elif(command_id == 7):
            expanded_cmds.extend([(command_id, -1, -1)]*command_count)
        else:
            print("ERROR: Unknown command encountered.")
        cmds_pc += 1
    return expanded_cmds

def parse_point(feature: vt_proto.Tile.Feature, keys, values, xtile, ytile, zoom, extent):
    geometry_cmds = feature.geometry
    geom_pc = 0

    acquired_points = []

    cX = 0
    cY = 0

    while(geom_pc < len(geometry_cmds)):
        command_integer = geometry_cmds[geom_pc]
        command_id = command_integer & 0x07
        command_count = command_integer >> 3
        
        if(command_id == 1):
            for i in range(command_count):
                geom_pc += 1
                dX = unzigzag_coords(geometry_cmds[geom_pc])
                geom_pc += 1
                dY = unzigzag_coords(geometry_cmds[geom_pc])
                x = cX + dX
                y = cY + dY
                acquired_points.append(offset_to_latlon(xtile, ytile, zoom, x, y, extent))
                cX = x
                cY = y
        else:
            print("ERROR: Unknown command encountered in Point feature.")
        geom_pc += 1
    
    tags = feature.tags
    tag_pairs = [(tags[i*2], tags[i*2+1]) for i in range(len(tags)//2)]
    properties = {keys[key]: values[value] for key, value in tag_pairs}

    if(len(acquired_points) == 1):
        return Point(
            coordinates = acquired_points[0],
            properties = properties
        )
    else:
        return MultiPoint(
            coordinates=acquired_points,
            properties=properties
        )

def parse_linestring(feature: vt_proto.Tile.Feature, keys, values, xtile, ytile, zoom, extent):
    geometry_cmds = feature.geometry

    acquired_lines = []

    cX = 0
    cY = 0

    all_commands = expand_commands(geometry_cmds)

    split_commands = []
    for command in all_commands:
        if(command[0] == 1):
            split_commands.append([command])
        elif(command[0] == 2):
            split_commands[-1].extend([command])
    
    for lines in split_commands:
        line_coords = []
        for command_id, x, y in lines:
            if(command_id == 1 or command_id == 2):
                dX = unzigzag_coords(x)
                dY = unzigzag_coords(y)
                x = cX + dX
                y = cY + dY
                line_coords.append(offset_to_latlon(xtile, ytile, zoom, x, y, extent))
                cX = x
                cY = y
            else:
                print("ERROR: ClosePath command found in LineString feature, unexpected.")
                line_coords.append(line_coords[0])
        acquired_lines.append(line_coords)
    
    tags = feature.tags
    tag_pairs = [(tags[i*2], tags[i*2+1]) for i in range(len(tags)//2)]
    properties = {keys[key]: values[value] for key, value in tag_pairs}

    if(len(acquired_lines) > 1):
        return MultiLineString(
            coordinates=acquired_lines,
            properties=properties
        )
    else:
        return LineString(
            coordinates=acquired_lines[0],
            properties=properties
        )

def parse_polygon(feature: vt_proto.Tile.Feature, keys, values, xtile, ytile, zoom, extent):
    geometry_cmds = feature.geometry

    acquired_polygons = []

    tags = feature.tags
    tag_pairs = [(tags[i*2], tags[i*2+1]) for i in range(len(tags)//2)]
    properties = {keys[key]: values[value] for key, value in tag_pairs}

    cX = 0
    cY = 0

    all_commands = expand_commands(geometry_cmds)

    split_commands = []
    for command in all_commands:
        if(command[0] == 1):
            split_commands.append([command])
        elif(command[0] == 2 or command[0] == 7):
            split_commands[-1].extend([command])
    
    for polygons in split_commands:
        poly_coords = []
        poly_raw_coords = []
        for command_id, x, y in polygons:
            if(command_id == 1 or command_id == 2):
                dX = unzigzag_coords(x)
                dY = unzigzag_coords(y)
                x = cX + dX
                y = cY + dY
                poly_coords.append(offset_to_latlon(xtile, ytile, zoom, x, y, extent))
                poly_raw_coords.append((x, y))
                cX = x
                cY = y
            else:
                poly_coords.append(poly_coords[0])
                poly_raw_coords.append(poly_raw_coords[0])
        
        area = 0
        if(poly_raw_coords[0] == poly_raw_coords[-1]):
            area = area_by_shoelace(poly_raw_coords[:-1])
        else:
            area = area_by_shoelace(poly_raw_coords)
        
        if (area >= 0):
            acquired_polygons.append((1, poly_coords))
        else:
            acquired_polygons.append((-1, poly_coords))

    # Divide into sequences of exterior and interior rings
    polygon_items = []
    for polygon in acquired_polygons:
        if(polygon[0] > 0):
            # Exterior ring
            polygon_items.append([polygon[1]])
        else:
            # Interior ring
            if(len(polygon_items) >= 1):
                polygon_items[-1].extend([polygon[1]])
            else:
                print("ERROR: Interior ring found before exterior ring.")
    
    if(len(polygon_items) > 1):
        return MultiPolygon(
            coordinates=polygon_items,
            properties=properties
        )
    else:
        return Polygon(
            coordinates=polygon_items[0],
            properties=properties
        )
    
def extract_layer(layer: vt_proto.Tile.Layer, xtile, ytile, zoom) -> Tuple[str, FeatureCollection]:
    layer_content = []

    keys = layer.keys
    values = []
    for value in layer.values:
        if(value.HasField('string_value')):
            values.append(value.string_value)
        elif(value.HasField('double_value')):
            values.append(value.double_value)
        elif(value.HasField('float_value')):
            values.append(value.float_value)
        elif(value.HasField('int_value')):
            values.append(value.int_value)
        elif(value.HasField('uint_value')):
            values.append(value.uint_value)
        elif(value.HasField('sint_value')):
            values.append(value.sint_value)
        elif(value.HasField('bool_value')):
            values.append(value.bool_value)
    
    for feature in layer.features:
        geom_type = feature.type

        if(geom_type == vt_proto.Tile.GeomType.POINT):
            layer_content.append(Feature(geometry=parse_point(feature, keys, values, xtile, ytile, zoom, layer.extent)))
        elif(geom_type == vt_proto.Tile.GeomType.LINESTRING):
            layer_content.append(Feature(geometry=parse_linestring(feature, keys, values, xtile, ytile, zoom, layer.extent)))
        elif(geom_type == vt_proto.Tile.GeomType.POLYGON):
            layer_content.append(Feature(geometry=parse_polygon(feature, keys, values, xtile, ytile, zoom, layer.extent)))
    return layer.name, FeatureCollection(layer_content)



import argparse
import asyncio
from .decoder.BytesDecoder import BytesDecoder
import json
import os
import sys
import re
from aiohttp_retry import RetryClient, ExponentialRetry
import geojson

pattern = re.compile("https://tiles.mapillary.com/maps/vtp/([^/]*)/2/(\\d+)/{x}/{y}.*")
fixed_pattern = re.compile("https://tiles.mapillary.com/maps/vtp/([^/]*)/2/(\\d+)/(\\d+)/(\\d+).*")

async def worker(url, client, output_dir, tile_name, json_indent, xtile, ytile, zoom, split_layers):
    print(f"Fetching tile {zoom}-{xtile}-{ytile}")
    async with client.get(url) as response:
        if(response.status == 429):
            asyncio.sleep(5)
        elif(response.status // 100 == 2):
            body = await response.read()
            decoder = BytesDecoder(xtile, ytile, zoom, body)
            result = decoder.decode()
            if(split_layers):
                for layer_name, layer_content in result.items():
                    output_filename = os.path.join(output_dir, f"{tile_name}-{zoom}-{xtile}-{ytile}-{layer_name}.json")
                    print("Writing layer {} to {}".format(layer_name, output_filename))
                    with open(output_filename, 'w') as f:
                        geojson.dump(layer_content, f, indent=json_indent if json_indent > 0 else None)
            else:
                output_filename = os.path.join(output_dir, f"{tile_name}-{zoom}-{xtile}-{ytile}.json")
                with open(output_filename, 'w') as f:
                    json.dump(result, f, indent=json_indent if json_indent > 0 else None)
                print(f"Writing all layers to {output_filename}")
        elif(response.status // 100 == 5):
            asyncio.sleep(10)

async def run(url: str, start_x: int, start_y: int, end_x: int, end_y: int, output_dir, json_indent, split_layers):
    os.makedirs(output_dir, exist_ok=True)

    match = re.match(pattern, url)
    if(match is None):
        print("URL does not match Mapillary tile request pattern.")
        return
    
    tile_name = match[1]
    zoom = int(match[2])
    async with RetryClient(raise_for_status=False, retry_options=ExponentialRetry(attempts=5, start_timeout=0.5, max_timeout=10)) as client:
        for x in range(start_x, end_x + 1):
            for y in range(start_y, end_y + 1):
                tile_url = url.replace('{x}', str(x)).replace('{y}', str(y))
                output_filename = os.path.join(output_dir, f"{tile_name}-{zoom}-{x}-{y}.json")
                await worker(tile_url, client, output_dir, tile_name, json_indent, x, y, zoom, split_layers)

async def run_fixed(url: str, output_dir: str, output_filename: str, json_indent, split_layers):
    match = re.match(fixed_pattern, url)
    if(match is None):
        print("URL does not match Mapillary tile request pattern.")
        return
    
    tile_name = match[1]
    zoom = int(match[2])
    xtile = int(match[3])
    ytile = int(match[4])

    async with RetryClient(raise_for_status=False, retry_options=ExponentialRetry(attempts=5, start_timeout=0.5, max_timeout=10)) as client:
        async with client.get(url) as response:
            if(response.status == 429):
                asyncio.sleep(5)
            elif(response.status // 100 == 2):
                body = await response.read()
                decoder = BytesDecoder(xtile, ytile, zoom, body)
                result = decoder.decode()
                if(split_layers):
                    for layer_name, layer_content in result.items():
                        if(output_filename is not None):
                            if(output_filename.endswith(".json")):
                                layer_filename = output_filename.replace(".json", f"-{layer_name}.json")
                            else:
                                layer_filename = output_filename + f"-{layer_name}.json"
                        else:
                            layer_filename = os.path.join(output_dir, f"{tile_name}-{zoom}-{xtile}-{ytile}-{layer_name}.json")

                        print("Writing layer {} to {}".format(layer_name, layer_filename))
                        with open(layer_filename, 'w') as f:
                            geojson.dump(layer_content, f, indent=json_indent if json_indent > 0 else None)
                else:
                    if(output_filename is None):
                        layers_filename = os.path.join(output_dir, f"{tile_name}-{zoom}-{xtile}-{ytile}.json")
                    else:
                        layers_filename = output_filename
                    with open(layers_filename, 'w') as f:
                        json.dump(result, f, indent=json_indent if json_indent > 0 else None)
                    print(f"Writing all layers to {layers_filename}")
            elif(response.status // 100 == 5):
                asyncio.sleep(10)

def main():
    parser = argparse.ArgumentParser(description="Fetch multiple tiles from mapillary.com and convert to GeoJSON.")
    parser.add_argument("--url", dest = "url", help="URL template of tiles to fetch. {x} and {y} in the template will be replaced.", required=True)
    parser.add_argument("--start-x", dest = "start_x", help="X coordinate of first tile", required=False, type=int)
    parser.add_argument("--start-y", dest = "start_y", help="Y coordinate of first tile", required=False, type=int)
    parser.add_argument("--end-x", dest = "end_x", help="X coordinate of last tile (inclusive)", required=False, type=int)
    parser.add_argument("--end-y", dest = "end_y", help="Y coordinate of last tile (inclusive)", required=False, type=int)

    parser.add_argument("--json-indent", dest="json_indent", help="JSON file indentation. 0 or negative numbers generate dense JSON file.", default=0, type=int, required = False)
    parser.add_argument("--output-dir", dest = "output_dir", help="Output directory", required=False)
    parser.add_argument("--split-layers", dest = "split_layers", help="Split layers into separate GeoJSON files. Outputs Pure GeoJSON.", action="store_true", default=False)
    parser.add_argument("--output", dest = "output", help="Output file", required=False)

    args = parser.parse_args()

    if(sys.platform.lower().startswith("win")):
        try:
            asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
        except:
            pass

    if(args.output_dir is None and args.output is None):
        print("Must specify either --output or --output-dir")
        exit(1)
    
    if(args.output_dir is not None):
        os.makedirs(args.output_dir, exist_ok=True)
    
    if(args.start_x is None or args.start_y is None or args.end_x is None or args.end_y is None):
        asyncio.run(run_fixed(args.url, args.output_dir, args.output, args.json_indent, args.split_layers))
        exit(0)
    
    if(args.start_x > args.end_x):
        print("Start X coordinate must be smaller than end X coordinate.")
        exit(1)
    elif(args.start_y > args.end_y):
        print("Start Y coordinate must be smaller than end Y coordinate.")
        exit(1)
    else:
        asyncio.run(run(args.url, args.start_x, args.start_y, args.end_x, args.end_y, args.output_dir, args.json_indent, args.split_layers))

if __name__ == '__main__':
    main()


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: vector_tile.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x11vector_tile.proto\x12\x0bvector_tile\"\xc0\x04\n\x04Tile\x12\'\n\x06layers\x18\x03 \x03(\x0b\x32\x17.vector_tile.Tile.Layer\x1a\xa1\x01\n\x05Value\x12\x14\n\x0cstring_value\x18\x01 \x01(\t\x12\x13\n\x0b\x66loat_value\x18\x02 \x01(\x02\x12\x14\n\x0c\x64ouble_value\x18\x03 \x01(\x01\x12\x11\n\tint_value\x18\x04 \x01(\x03\x12\x12\n\nuint_value\x18\x05 \x01(\x04\x12\x12\n\nsint_value\x18\x06 \x01(\x12\x12\x12\n\nbool_value\x18\x07 \x01(\x08*\x08\x08\x08\x10\x80\x80\x80\x80\x02\x1as\n\x07\x46\x65\x61ture\x12\r\n\x02id\x18\x01 \x01(\x04:\x01\x30\x12\x10\n\x04tags\x18\x02 \x03(\rB\x02\x10\x01\x12\x31\n\x04type\x18\x03 \x01(\x0e\x32\x1a.vector_tile.Tile.GeomType:\x07UNKNOWN\x12\x14\n\x08geometry\x18\x04 \x03(\rB\x02\x10\x01\x1a\xad\x01\n\x05Layer\x12\x12\n\x07version\x18\x0f \x02(\r:\x01\x31\x12\x0c\n\x04name\x18\x01 \x02(\t\x12+\n\x08\x66\x65\x61tures\x18\x02 \x03(\x0b\x32\x19.vector_tile.Tile.Feature\x12\x0c\n\x04keys\x18\x03 \x03(\t\x12\'\n\x06values\x18\x04 \x03(\x0b\x32\x17.vector_tile.Tile.Value\x12\x14\n\x06\x65xtent\x18\x05 \x01(\r:\x04\x34\x30\x39\x36*\x08\x08\x10\x10\x80\x80\x80\x80\x02\"?\n\x08GeomType\x12\x0b\n\x07UNKNOWN\x10\x00\x12\t\n\x05POINT\x10\x01\x12\x0e\n\nLINESTRING\x10\x02\x12\x0b\n\x07POLYGON\x10\x03*\x05\x08\x10\x10\x80@B\x02H\x03')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'vector_tile_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'H\003'
  _TILE_FEATURE.fields_by_name['tags']._options = None
  _TILE_FEATURE.fields_by_name['tags']._serialized_options = b'\020\001'
  _TILE_FEATURE.fields_by_name['geometry']._options = None
  _TILE_FEATURE.fields_by_name['geometry']._serialized_options = b'\020\001'
  _TILE._serialized_start=35
  _TILE._serialized_end=611
  _TILE_VALUE._serialized_start=85
  _TILE_VALUE._serialized_end=246
  _TILE_FEATURE._serialized_start=248
  _TILE_FEATURE._serialized_end=363
  _TILE_LAYER._serialized_start=366
  _TILE_LAYER._serialized_end=539
  _TILE_GEOMTYPE._serialized_start=541
  _TILE_GEOMTYPE._serialized_end=604
# @@protoc_insertion_point(module_scope)



from time import sleep
from pyparsing import line
from . import vector_tile_pb2 as vt_proto
from geojson import Feature, Point, FeatureCollection, LineString, MultiLineString, MultiPolygon, Polygon, MultiPoint
from typing import Dict, Tuple, List
import math
from .utils import unzigzag_coords, expand_commands, area_by_shoelace
import multiprocessing
from concurrent.futures import ThreadPoolExecutor

class LayerDecoder:
    def __init__(self, filedecoder, layer: vt_proto.Tile.Layer):
        "Initialize a Layer decoder from a file decoder."
        self.extent = layer.extent
        self.layer = layer
        self.decoded = None

        self.xtile = filedecoder.xtile
        self.ytile = filedecoder.ytile
        self.zoom = filedecoder.zoom

        self.keys = []
        self.values = []

    # Utility Functions.
    def offset_to_latlon(self, xoffset, yoffset) -> Tuple[float, float]:
        x_tiled_coord = self.xtile * self.extent + xoffset
        y_tiled_coord = self.ytile * self.extent + yoffset

        lon = x_tiled_coord * 360 / (self.extent * 2 ** self.zoom) - 180

        lat_noncoerce = 180 - y_tiled_coord * 360. / (self.extent * 2 ** self.zoom)
        lat = 360. / math.pi * math.atan(math.exp(lat_noncoerce * math.pi / 180)) - 90
        return (lon, lat)
    
    def extract_properties(self) -> None:
        self.keys = self.layer.keys
        for value in self.layer.values:
            if(value.HasField('string_value')):
                self.values.append(value.string_value)
            elif(value.HasField('double_value')):
                self.values.append(value.double_value)
            elif(value.HasField('float_value')):
                self.values.append(value.float_value)
            elif(value.HasField('int_value')):
                self.values.append(value.int_value)
            elif(value.HasField('uint_value')):
                self.values.append(value.uint_value)
            elif(value.HasField('sint_value')):
                self.values.append(value.sint_value)
            elif(value.HasField('bool_value')):
                self.values.append(value.bool_value)
    
    # Parser functions.
    def parse_point(self, feature: vt_proto.Tile.Feature):
        geometry_cmds = feature.geometry
        geom_pc = 0

        acquired_points = []

        cX = 0
        cY = 0

        while(geom_pc < len(geometry_cmds)):
            command_integer = geometry_cmds[geom_pc]
            command_id = command_integer & 0x07
            command_count = command_integer >> 3
            
            if(command_id == 1):
                for i in range(command_count):
                    geom_pc += 1
                    dX = unzigzag_coords(geometry_cmds[geom_pc])
                    geom_pc += 1
                    dY = unzigzag_coords(geometry_cmds[geom_pc])
                    x = cX + dX
                    y = cY + dY
                    acquired_points.append(self.offset_to_latlon(x, y))
                    cX = x
                    cY = y
            else:
                print("ERROR: Unexpected command encountered in Point feature. Ignored.")
            geom_pc += 1
        
        tags = feature.tags
        tag_pairs = [(tags[i*2], tags[i*2+1]) for i in range(len(tags)//2)]
        properties = {self.keys[key]: self.values[value] for key, value in tag_pairs}

        if(len(acquired_points) == 1):
            return Point(
                coordinates = acquired_points[0],
                properties = properties
            )
        else:
            return MultiPoint(
                coordinates=acquired_points,
                properties=properties
            )
    
    def parse_linestring(self, feature: vt_proto.Tile.Feature):
        geometry_cmds = feature.geometry
        acquired_lines = []

        cX = 0
        cY = 0

        all_commands = expand_commands(geometry_cmds)
        split_commands = []
        for command in all_commands:
            if(command[0] == 1):
                split_commands.append([command])
            elif(command[0] == 2):
                split_commands[-1].extend([command])
        
        for lines in split_commands:
            line_coords = []
            for command_id, x, y in lines:
                if(command_id == 1 or command_id == 2):
                    dX = unzigzag_coords(x)
                    dY = unzigzag_coords(y)
                    x = cX + dX
                    y = cY + dY
                    line_coords.append(self.offset_to_latlon(x, y))
                    cX = x
                    cY = y
                else:
                    print("ERROR: ClosePath command found in LineString feature. This is likely an error, closing the linestring feature as a best-effort guess.")
                    line_coords.append(line_coords[0])
            acquired_lines.append(line_coords)
        
        tags = feature.tags
        tag_pairs = [(tags[i*2], tags[i*2+1]) for i in range(len(tags)//2)]
        properties = {self.keys[key]: self.values[value] for key, value in tag_pairs}

        if(len(acquired_lines) > 1):
            return MultiLineString(
                coordinates=acquired_lines,
                properties=properties
            )
        else:
            return LineString(
                coordinates=acquired_lines[0],
                properties=properties
            )

    def parse_polygon(self, feature: vt_proto.Tile.Feature):
        geometry_cmds = feature.geometry

        acquired_polygons = []

        tags = feature.tags
        tag_pairs = [(tags[i*2], tags[i*2+1]) for i in range(len(tags)//2)]
        properties = {self.keys[key]: self.values[value] for key, value in tag_pairs}

        cX = 0
        cY = 0

        all_commands = expand_commands(geometry_cmds)

        split_commands = []
        for command in all_commands:
            if(command[0] == 1):
                split_commands.append([command])
            elif(command[0] == 2 or command[0] == 7):
                split_commands[-1].extend([command])
        
        # Sanity check: All polygons should have a start command MoveTo and an end command ClosePath.
        for command in split_commands:
            if(command[0][0] != 1):
                print("WARN: Polygon command does not start with MoveTo. There might be unknown consequences.")
            if(command[-1][0] != 7):
                print("WARN: Polygon command does not end with ClosePath. There might be unknown consequences.")
        
        for polygons in split_commands:
            poly_coords = []
            poly_raw_coords = []
            for command_id, x, y in polygons:
                if(command_id == 1 or command_id == 2):
                    dX = unzigzag_coords(x)
                    dY = unzigzag_coords(y)
                    x = cX + dX
                    y = cY + dY
                    poly_coords.append(self.offset_to_latlon(x, y))
                    poly_raw_coords.append((x, y))
                    cX = x
                    cY = y
                else:
                    poly_coords.append(poly_coords[0])
                    poly_raw_coords.append(poly_raw_coords[0])
            
            area = 0
            if(poly_raw_coords[0] == poly_raw_coords[-1]):
                area = area_by_shoelace(poly_raw_coords[:-1])
            else:
                area = area_by_shoelace(poly_raw_coords)
            
            if (area >= 0):
                acquired_polygons.append((1, poly_coords))
            else:
                acquired_polygons.append((-1, poly_coords))

        # Divide into sequences of exterior and interior rings
        polygon_items = []
        for polygon in acquired_polygons:
            if(polygon[0] > 0):
                # Exterior ring
                polygon_items.append([polygon[1]])
            else:
                # Interior ring
                if(len(polygon_items) >= 1):
                    polygon_items[-1].extend([polygon[1]])
                else:
                    print("ERROR: Interior ring found before exterior ring. The interior ring will be ignored.")
        
        if(len(polygon_items) > 1):
            return MultiPolygon(
                coordinates=polygon_items,
                properties=properties
            )
        else:
            return Polygon(
                coordinates=polygon_items[0],
                properties=properties
            )
    
    def parse_feature(self, feature: vt_proto.Tile.Feature) -> Feature:
        geom_type = feature.type
        if(geom_type == vt_proto.Tile.GeomType.POINT):
            return Feature(geometry=self.parse_point(feature))
        elif(geom_type == vt_proto.Tile.GeomType.LINESTRING):
            return Feature(geometry=self.parse_linestring(feature))
        elif(geom_type == vt_proto.Tile.GeomType.POLYGON):
            return Feature(geometry=self.parse_polygon(feature))

    def decode(self) -> Tuple[str, FeatureCollection]:
        if(self.decoded is None):
            self.extract_properties()

            layer_content = []

            num_cpus = multiprocessing.cpu_count()
            if(num_cpus == None):
                num_cpus = 4

            features_list = list(self.layer.features)

            with ThreadPoolExecutor(max_workers=num_cpus) as executor:
                layer_content = list(executor.map(self.parse_feature, features_list))
                executor.shutdown(wait=True)
            
            self.decoded = self.layer.name, FeatureCollection(layer_content)
        
        return self.decoded


from pickletools import bytes_types
from pyparsing import line
from .LayerDecoder import LayerDecoder
from . import vector_tile_pb2 as vt_proto
from geojson import Feature, Point, FeatureCollection, LineString, MultiLineString, MultiPolygon, Polygon, MultiPoint
from typing import Dict, Tuple, List
import multiprocessing
from concurrent.futures import ThreadPoolExecutor

class BytesDecoder:
    def __init__(self, xtile: int, ytile: int, zoom: int, bytes: bytes):
        "Load the file decoder with a file. Decoding has not started."
        self.xtile = xtile
        self.ytile = ytile
        self.zoom = zoom
        self.bytes = bytes
        self.decoded = None
    
    def read_protobuf(self) -> vt_proto.Tile:
        return vt_proto.Tile.FromString(self.bytes)
    
    def decode_layer(self, layer: vt_proto.Tile.Layer) -> Tuple[str, FeatureCollection]:
        layer_decoder = LayerDecoder(self, layer)
        layer_name, layer_content = layer_decoder.decode()
        return (layer_name, layer_content)
    
    def decode(self):
        "Perform the decoding of the file. Multiple calls will not re-decode."
        if(self.decoded is None):
            tile = self.read_protobuf()
            layers = tile.layers
            self.decoded = dict()

            with ThreadPoolExecutor(max_workers=multiprocessing.cpu_count()) as executor:
                decoded_layers = executor.map(self.decode_layer, layers)
                executor.shutdown(wait=True)
                for layer_name, layer_content in decoded_layers:
                    self.decoded[layer_name] = layer_content

        return self.decoded


import math
from typing import Dict, Tuple, List

def area_by_shoelace(points:  List[Tuple[float, float]]) -> float:
    "Assumes x,y points go around the polygon in one direction"
    x, y = zip(*points)
    return (sum(i * j for i, j in zip(x,             y[1:] + y[:1]))
           -sum(i * j for i, j in zip(x[1:] + x[:1], y            ))) / 2
               
def num2deg(xtile: int, ytile: int, zoom: int) -> Tuple[float, float]:
  n = 2.0 ** zoom
  lon_deg = xtile / n * 360.0 - 180.0
  lat_rad = math.atan(math.sinh(math.pi * (1 - 2 * ytile / n)))
  lat_deg = math.degrees(lat_rad)
  return (lat_deg, lon_deg)

def unzigzag_coords(value: int) -> int:
    return (value >> 1) ^ (-(value & 1))

def expand_commands(cmds: List[int]) -> List[Tuple]:
    expanded_cmds = []
    cmds_pc = 0
    while(cmds_pc < len(cmds)):
        command_integer = cmds[cmds_pc]
        command_id = command_integer & 0x07
        command_count = command_integer >> 3
        if(command_id == 1 or command_id == 2):
            for i in range(command_count):
                cmds_pc += 1
                param1 = cmds[cmds_pc]
                cmds_pc += 1
                param2 = cmds[cmds_pc]
                expanded_cmds.append((command_id, param1, param2))
        elif(command_id == 7):
            expanded_cmds.extend([(command_id, -1, -1)]*command_count)
        else:
            print("ERROR: Unknown command encountered.")
        cmds_pc += 1
    return expanded_cmds